package db;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.clickhouse.client.api.Client;
import com.clickhouse.client.api.data_formats.ClickHouseBinaryFormatReader;
import com.clickhouse.client.api.enums.Protocol;
import com.clickhouse.client.api.metadata.TableSchema;
import com.clickhouse.client.api.query.QueryResponse;
import com.clickhouse.client.api.query.QuerySettings;
import db.dto.StckPointVal;
import db.page.MultiPointRangePagingQuery;
import db.page.RangePagingQuery;
import org.apache.commons.lang3.StringUtils;
import org.javatuples.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.math.BigDecimal;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.util.*;
import java.util.stream.Collectors;

// TODO: Add log support

/// 数据库 API 接口类
public class DBApiEntry implements AutoCloseable {
    private static final Logger LOG = LoggerFactory.getLogger(DBApiEntry.class);

    private static final double VALUE_INVALID = Double.NaN;
    //    private static final double VALUE_INVALID = -10000;
    private final static String TABLE_NAME = "tsdb_cpp";
    private final Client client;
    private String host;
    private int port;
    private int insertPort;
    private int selectPort;
    private boolean addNullValue = true;
    private int splitSize = 1000000;
    private int oneThreadDataSize = 100000;
    private String metricNamePrefix = "";
    private boolean trimTagName = true;
    private int splitSizeRealTime = 20;
    private int nameSpace = 1;
    private boolean isUseLB = false;
    private String token = null;
    //按步长取值的超时时间，默认为10分钟
    private long lookBack = 10L * 60 * 1000_000_000;
    private String currentDatabase = "default";
    private String currentTable = TABLE_NAME;

    /**
     * 初始化数据驱动构造方法
     *
     * @throws IOException
     */
    public DBApiEntry(String addressHost, int publicPort) throws IOException {
//        this.host = addressHost;
//        this.port = publicPort;
//        this.insertPort = this.port + 1;
//        this.selectPort = this.port + 2;
        client = new Client.Builder()
                .addEndpoint(Protocol.HTTP, addressHost, publicPort, false)
                .setUsername("default")
                .setPassword("")
                .compressClientRequest(true)
                .compressServerResponse(true)
                .build();

        // Init ClickHouse client
        TableSchema schema = client.getTableSchema(makeTableName(), currentDatabase);
        client.register(StckPointVal.class, schema);
    }

    public DBApiEntry(String addressHost, int publicPort, String token) throws IOException {
//        this.host = addressHost;
//        this.port = publicPort;
//        this.token = token;
//        this.insertPort = this.port + 1;
//        this.selectPort = this.port + 2;
        this(addressHost, publicPort);
    }

    /**
     * 静态方法初始化时序数据源驱动类
     *
     * @return DBApiEntry
     */
    public static DBApiEntry initApiEntry(String addressHost, int publicPort) {
        try {
            return new DBApiEntry(addressHost, publicPort);
        } catch (Exception e) {
            LOG.error("initApiEntry error: ", e);
            return null;
        }
    }

    public static DBApiEntry initApiEntry(String addressHost, int publicPort, String token) {
        return initApiEntry(addressHost, publicPort);
    }

    /**
     * 按指定大小，分隔集合，将集合按规定个数分为n个部分
     *
     * @param dbValList
     * @return
     * @date
     */
    protected static List<List<DBVal>> splitDataList(List<DBVal> dbValList, int splitSize) {
        if (dbValList == null || dbValList.size() == 0 || splitSize < 1) {
            return new ArrayList<>();
        }

        List<List<DBVal>> resultList = new ArrayList<>();

        if (dbValList.size() <= splitSize) {
            resultList.add(dbValList);
            return resultList;
        }

        int size = dbValList.size();
        int count = (size + splitSize - 1) / splitSize;

        for (int i = 0; i < count; i++) {
            List<DBVal> subList = dbValList.subList(i * splitSize, (Math.min((i + 1) * splitSize, size)));
            resultList.add(subList);
        }
        return resultList;
    }

    protected static List<PointVals> splitPointVals(PointVals pointVals, int splitSize) {
        if (pointVals == null || pointVals.isEmpty() || splitSize < 1) {
            return null;
        }

        List<PointVals> resultList = new ArrayList<>();

        if (pointVals.getCount() <= splitSize) {
            resultList.add(pointVals);
            return resultList;
        }

        int size = pointVals.getCount();
        int count = (size + splitSize - 1) / splitSize;

        for (int i = 0; i < count; i++) {
            PointVals subPointVals = pointVals.subPointVals(i * splitSize, (Math.min((i + 1) * splitSize, size)));
            resultList.add(subPointVals);
        }
        return resultList;
    }

    /**
     * 配套插入大型数据集的json生成
     * 单点多条数据
     */
    protected static String convertPointValsToImportJson(PointVals pointVals, String metricNamePrefix) {
        JSONObject metrics = new JSONObject();
        JSONObject metric = new JSONObject();
        JSONArray values = new JSONArray();
        JSONArray timestamps = new JSONArray();
        metric.put("__name__", pointVals.getFinalMetricName(metricNamePrefix));
        metric.putAll(pointVals.getTags());
        for (int index = 0; index < pointVals.getCount(); index++) {
            values.add(pointVals.getValue(index));
            timestamps.add(pointVals.getUtcTime(index));
        }
        metrics.put("metric", metric);
        metrics.put("timestamps", timestamps);
        metrics.put("values", values);
        return metrics.toJSONString();
    }

    public static String convertDBValToImportJson(String metricNamePrefix, DBVal val) {
        JSONObject metrics = new JSONObject();
        JSONObject metric = new JSONObject();
        JSONArray values = new JSONArray();
        JSONArray timestamps = new JSONArray();
        metric.put("__name__", val.getFinalMetricName(metricNamePrefix));
        metric.putAll(val.getTags());
        values.add(val.getValue());
        timestamps.add(val.getUtcTimeMilliSeconds());
        metrics.put("metric", metric);
        metrics.put("timestamps", timestamps);
        metrics.put("values", values);
        return metrics.toJSONString();
    }

    @Override
    public void close() throws Exception {
        client.close();
    }

    public void setUseLB(boolean useLB) {
        isUseLB = useLB;

        if (this.isUseLB) {
            this.insertPort = this.port;
            this.selectPort = this.port;
        } else {
            this.insertPort = this.port + 1;
            this.selectPort = this.port + 2;
        }
    }

    public String getToken() {
        return token;
    }

    public void setToken(String token) {
        this.token = token;
    }

    public int getNameSpace() {
        return nameSpace;
    }

    public void setNameSpace(int nameSpace) {
        this.nameSpace = nameSpace;
    }

    public String getMetricNamePrefix() {
        return metricNamePrefix;
    }

    public void setMetricNamePrefix(String metricNamePrefix) {
        this.metricNamePrefix = metricNamePrefix;
    }

    public long getLookBack() {
        return lookBack;
    }

    public void setLookBack(long lookBack) {
        this.lookBack = lookBack;
    }

    /**
     * 是否进行 trim 操作
     *
     * @return
     */
    public boolean isTrimTagName() {
        return trimTagName;
    }

    public boolean isAddNullValue() {
        return addNullValue;
    }

    /**
     * 点名合法性检查函数，强化版
     *
     * @return
     * @date
     */

    protected String trimStrong(String tagName) {
        try {
            String regEx = "[\n`~!@#$%^&*(){}':;',\\[\\].<>/?~！@#￥%……&*（）+{}【】‘；：”“’。， 、？] \\s";
            tagName = tagName.replaceAll(regEx, "");
        } catch (Exception e) {
            e.printStackTrace();
        }
        return tagName;
    }

    protected List<DBVal> addNullValues(List<DBVal> dbValList, String tagName, long start, long end, int step) {
        Hashtable<Long, DBVal> dbValHashtable = new Hashtable<>();
        for (DBVal dbVal : dbValList) {
            dbValHashtable.put(dbVal.getUtcTime(), dbVal);
        }
        ArrayList<DBVal> result = new ArrayList<>();

        for (long time = start; time <= end; time += step) {
            DBVal dbVal = dbValHashtable.get(time);
            if (dbVal == null) {
                dbVal = new DBVal();
                dbVal.setUtcTime(time);
                dbVal.setMetricName(tagName);
                dbVal.setValue(VALUE_INVALID);
            }
            result.add(dbVal);
        }
        return result;
    }

    /**
     * @date 写入多点多条数据
     * 被sendMultiPoint调用
     */
    protected int splitAndInsertMultiPoint(List<DBVal> dbVals) {
        if (dbVals == null || dbVals.isEmpty()) {
            return 0;
        }

        int insertedCount = 0;
        try {
            List<List<DBVal>> dbValList = splitDataList(dbVals, this.splitSize);
            for (List<DBVal> subList : dbValList) {
                this.insertDBValsToDB(subList);
                insertedCount += subList.size();
            }
        } catch (Exception e) {
            LOG.error("splitAndInsertMultiPoint error: ", e);
        }
        return insertedCount;

//        try {
//            int counts = 0;
//            StringBuilder sentence = new StringBuilder(50000);
//            String item = "";
//            for (DBVal val : dbVals) {
//                item = convertDBValToImportJson(this.getMetricNamePrefix(), val);
//                sentence.append(item).append("\n");
//                counts++;
//                if (counts > splitSize) {
//                    insertByHttp(sentence.toString());
//                    counts = 0;
//                    sentence.setLength(0);
//                }
//            }
//            //将不够长的一起发出去
//            if (sentence.length() != 0) {
//                insertByHttp(sentence.toString());
//            }
////            sentence.setLength(0);
//        } catch (Exception e) {
//            e.printStackTrace();
//            return -1;
//        }
//        return 1;
    }

    /**
     * 本方法用于数据插入
     *
     * @param sentences String类型,json格式的数据,组装好之后直接作为参数传入
     * @return void
     * @date sendMultiPoint最终调用
     */
    protected void insertByHttp(String sentences) {
        try {
            String sendDataUrl = String.format("http://%s:%d/insert/%d/%s/api/v1/import", host, insertPort, nameSpace, Constants.ROUTE_PREFIX);

            URL realUrl = new URL(sendDataUrl);
            URLConnection conn = realUrl.openConnection();
            conn.setRequestProperty("accept", "*/*");
            conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
            conn.setRequestProperty("host", "application/x-www-form-urlencoded");
            this.configHttpHeaderToken(conn);
            conn.setDoOutput(true);
            conn.setDoInput(true);
            DataOutputStream dos = new DataOutputStream(conn.getOutputStream());
            dos.write(sentences.getBytes());
            dos.flush();
            BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    protected void configHttpHeaderToken(URLConnection conn) {
        if (this.getToken() != null) {
            conn.setRequestProperty("Authorization", "Bearer " + this.getToken());
        }
    }

    public String execSql(String sql) {
        StringBuffer result = new StringBuffer();
        try {
            String encodesSql = sql.replace("[", "%5B").replace("]", "%5D").replace(" ", "%20").replace("{", "%7B").replace("}", "%7D").replace("\"", "%22");

            String sqlUrl = String.format("http://%s:%d/sql/%d/api/v1/sql?sql=%s", host, selectPort, nameSpace, encodesSql);

            URL realUrl = new URL(sqlUrl);
            URLConnection conn = realUrl.openConnection();
            conn.setRequestProperty("accept", "*/*");
            conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
            conn.setRequestProperty("host", "application/x-www-form-urlencoded");
            this.configHttpHeaderToken(conn);
            conn.setDoOutput(true);
            conn.setDoInput(true);
            DataOutputStream dos = new DataOutputStream(conn.getOutputStream());
            dos.flush();

            BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));

            String line;
            while ((line = in.readLine()) != null) {
                result.append(line);
            }
            in.close();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        return result.toString();
    }

    public String execInsert(String sql) {
        StringBuffer result = new StringBuffer();
        try {
            String encodesSql = sql.replace("[", "%5B").replace("]", "%5D").replace(" ", "%20").replace("{", "%7B").replace("}", "%7D").replace("+", "%2B").replace("\"", "%22");
//            String encodesSql = sql;
            String sqlUrl = String.format("http://%s:%d/insert/%d/api/v1/sql?sql=%s", host, insertPort, nameSpace, encodesSql);

            URL realUrl = new URL(sqlUrl);
            URLConnection conn = realUrl.openConnection();
            conn.setRequestProperty("accept", "*/*");
            conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
            conn.setRequestProperty("host", "application/x-www-form-urlencoded");
            this.configHttpHeaderToken(conn);
            conn.setDoOutput(true);
            conn.setDoInput(true);
            DataOutputStream dos = new DataOutputStream(conn.getOutputStream());
            dos.flush();

            BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));

            String line;
            while ((line = in.readLine()) != null) {
                result.append(line);
            }
            in.close();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        return result.toString();
    }

    protected String makeTableName() {
        return String.format("%s.%s", currentDatabase, currentTable);
    }

    /// Write to ClickHouse
    protected void insertPointValsToDB(PointVals pointVals) {
        if (pointVals.isEmpty()) {
            return;
        }

        pointVals.getPoint().ensureValid();

        // Convert to ClickHousePointVal for insertion
        List<StckPointVal> pointValsList = new LinkedList<>();
        String metricName = pointVals.getMetricName();
        String pointName = pointVals.getPointName();
        HashMap<String, String> tags = new HashMap<>();
        for (String key : pointVals.getTags().keySet()) {
            if (StringUtils.equals(key, "pointName")) {
                continue;
            }
            tags.put(key, pointVals.getTags().get(key));
        }
        for (int i = 0; i < pointVals.getCount(); i++) {
            StckPointVal stckPointVal = new StckPointVal();
            stckPointVal.setMetricName(metricName);
            stckPointVal.setPointName(pointName);
            stckPointVal.setTags(tags);
            stckPointVal.setNanoseconds(pointVals.getUtcTime(i));
            stckPointVal.setValue(pointVals.getValue(i));
            pointValsList.add(stckPointVal);
        }
        try {
            client.insert(makeTableName(), pointValsList).get();
        } catch (Exception e) {
            LOG.error("insertPointValsToDb error: ", e);
        }
    }

    /// Write to ClickHouse
    protected void insertDBValsToDB(List<DBVal> dbVals) {
        if (dbVals == null || dbVals.isEmpty()) {
            return;
        }

        List<StckPointVal> pointValsList = new LinkedList<>();
        for (DBVal dbVal : dbVals) {
            StckPointVal stckPointVal = new StckPointVal();
            stckPointVal.setMetricName(dbVal.getMetricName());
            stckPointVal.setPointName(dbVal.getPointName());
            HashMap<String, String> tags = new HashMap<>();
            for (String key : dbVal.getTags().keySet()) {
                if (StringUtils.equals(key, "pointName")) {
                    continue;
                }
                tags.put(key, dbVal.getTags().get(key));
            }
            stckPointVal.setTags(tags);
            stckPointVal.setNanoseconds(dbVal.getUtcTimeMilliSeconds());
            stckPointVal.setValue(dbVal.getValue());
            pointValsList.add(stckPointVal);
        }
        try {
            client.insert(makeTableName(), pointValsList).get();
        } catch (Exception e) {
            LOG.error("insertDBValsToDB error: ", e);
        }
    }

    /// Read from ClickHouse
    protected PointVals selectPointValsFromDB(Point point, long tsStart, long tsEnd, boolean reversed, long limit) {
        Map<String, Object> queryParams = new HashMap<>();
        queryParams.put("metricName", point.getMetricName());
        queryParams.put("pointName", point.getPointName());
        queryParams.put("tsStart", tsStart);
        queryParams.put("tsEnd", tsEnd);
        String sql = "SELECT value, nanoseconds FROM " + makeTableName() +
                " WHERE metric_name = {metricName:String} AND point_name = {pointName:String}" +
                " AND nanoseconds >= {tsStart:Int64} AND nanoseconds <= {tsEnd:Int64}" +
                " ORDER BY (nanoseconds)";
        if (reversed) {
            sql += " DESC";
        }
        if (limit > 0) {
            sql += " LIMIT {limit:Int64}";
            queryParams.put("limit", limit);
        }
        try (QueryResponse response = client.query(sql, queryParams, new QuerySettings()).get()) {
            ClickHouseBinaryFormatReader reader = client.newBinaryFormatReader(response);
            List<PointVals> result = new ArrayList<>();
            int resultCount = (int) response.getResultRows();
            long[] timestamps = new long[resultCount];
            double[] values = new double[resultCount];
            int curReadPos = 0;
            while (reader.hasNext()) {
                reader.next();
                values[curReadPos] = reader.getDouble(1);
                timestamps[curReadPos] = reader.getLong(2);
                curReadPos++;
            }
            return new PointVals(point, resultCount, timestamps, values);
        } catch (Exception e) {
            LOG.error("selectPointValsFromDB error: ", e);
            return null;
        }
    }

    /// Read from ClickHouse
    protected List<PointVals> selectManyPointValsFromDB(List<Point> points, long tsStart, long tsEnd, boolean reversed, long limit) {
        // Deduplicate keys for values lookup
        Map<Pair<String, String>, Integer> pointsKeyMap = new HashMap<>();
        for (int i = 0; i < points.size(); i++) {
            pointsKeyMap.put(Pair.with(points.get(i).getMetricName(), points.get(i).getPointName()), i);
        }
        // Query values
        Map<String, Object> queryParams = new HashMap<>();
        queryParams.put("metricPointNames",
                pointsKeyMap
                        .keySet()
                        .stream()
                        .map(x -> String.format("('%s', '%s')", x.getValue0(), x.getValue1()))
                        .collect(Collectors.toList())
        );
        queryParams.put("tsStart", tsStart);
        queryParams.put("tsEnd", tsEnd);
        String sql = "SELECT metric_name, point_name, value, nanoseconds FROM " + makeTableName() +
                " WHERE (metric_name, point_name) in {metricPointNames:Array(Tuple(String, String))}" +
                " AND nanoseconds >= {tsStart:Int64} AND nanoseconds <= {tsEnd:Int64}" +
                " ORDER BY (metric_name, point_name, nanoseconds)";
        if (reversed) {
            sql += " DESC";
        }
        if (limit > 0) {
            sql += " LIMIT {limit:Int64} BY (metric_name, point_name)";
            queryParams.put("limit", limit);
        }
        try (QueryResponse response = client.query(sql, queryParams, new QuerySettings()).get()) {
            Map<Pair<String, String>, PointVals> pointsResultMap = new HashMap<>();
            ClickHouseBinaryFormatReader reader = client.newBinaryFormatReader(response);
            int resultCount = (int) response.getResultRows();
            long[] tsCache = new long[resultCount];
            double[] valueCache = new double[resultCount];
            Pair<String, String> curKey = Pair.with("", "");
            int curReadPos = 0;
            while (reader.hasNext()) {
                reader.next();
                String metricName = reader.getString(1);
                String pointName = reader.getString(2);
                Pair<String, String> nextKey = Pair.with(metricName, pointName);
                if (!curKey.equals(nextKey)) {
                    if (!curKey.getValue0().isEmpty()) {
                        PointVals pointVals = new PointVals(
                                points.get(pointsKeyMap.get(curKey)),
                                curReadPos,
                                Arrays.copyOf(tsCache, curReadPos),
                                Arrays.copyOf(valueCache, curReadPos)
                        );
                        pointsResultMap.put(curKey, pointVals);
                    }
                    curKey = nextKey;
                    curReadPos = 0;
                }
                valueCache[curReadPos] = reader.getDouble(3);
                tsCache[curReadPos] = reader.getLong(4);
                curReadPos++;
            }
            if (!curKey.getValue0().isEmpty() && curReadPos > 0) {
                PointVals pointVals = new PointVals(
                        points.get(pointsKeyMap.get(curKey)),
                        curReadPos,
                        Arrays.copyOf(tsCache, curReadPos),
                        Arrays.copyOf(valueCache, curReadPos)
                );
                pointsResultMap.put(curKey, pointVals);
            }
            // Build result list
            List<PointVals> result = new ArrayList<>();
            for (Point p : points) {
                Pair<String, String> key = Pair.with(p.getMetricName(), p.getPointName());
                PointVals pointVals = pointsResultMap.get(key);
                if (pointVals != null) {
                    result.add(pointVals);
                } else {
                    result.add(new PointVals(p, 0, new long[0], new double[0]));
                }
            }
            return result;
        } catch (Exception e) {
            LOG.error("selectManyPointValsFromDB error: ", e);
            return null;
        }
    }

    /// Read from ClickHouse
    ///
    /// 从数据库中读取指定时间范围内的数据，以指定步长取值（可看作在每个区间的端点处取值）
    ///
    /// @param step             步长，单位为纳秒
    /// @param lookBackDuration 从某点向前查找的最大时间范围（超出则标记无效），单位为纳秒
    protected PointVals selectPointValsFromDBWithStep(Point point, long tsStart, long tsEnd, long step, long lookBackDuration) {
        Map<String, Object> queryParams = new HashMap<>();
        queryParams.put("metricName", point.getMetricName());
        queryParams.put("pointName", point.getPointName());
        queryParams.put("tsStart", tsStart);
        queryParams.put("tsEnd", tsEnd);
        queryParams.put("step", step);
        queryParams.put("lookBackDuration", lookBackDuration);
        /*
SELECT a.*
FROM tsdb_cpp_memory AS a
INNER JOIN
(
    SELECT
        metric_name,
        point_name,
        max(nanoseconds) AS t
    FROM tsdb_cpp_memory
    GROUP BY (metric_name, point_name, intDiv(nanoseconds, 100))
) AS b ON a.nanoseconds = b.t
         */
        String sqlInner = "SELECT max(nanoseconds) AS t FROM " + makeTableName() +
                " WHERE metric_name = {metricName:String} AND point_name = {pointName:String}" +
                " AND nanoseconds >= ({tsStart:Int64} - {lookBackDuration:Int64}) AND nanoseconds <= {tsEnd:Int64}" +
                " GROUP BY (intDiv(nanoseconds - {tsStart:Int64}, {step:Int64}))";
        String sql = "SELECT a.value, a.nanoseconds FROM " + makeTableName() + " AS a" +
                " JOIN (" + sqlInner + ") AS b ON a.nanoseconds = b.t" +
                " WHERE metric_name = {metricName:String} AND point_name = {pointName:String}" +
                " AND nanoseconds >= ({tsStart:Int64} - {lookBackDuration:Int64}) AND nanoseconds <= {tsEnd:Int64}" +
                " ORDER BY (a.nanoseconds)";
        try (QueryResponse response = client.query(sql, queryParams, new QuerySettings()).get()) {
            ClickHouseBinaryFormatReader reader = client.newBinaryFormatReader(response);
            List<DBVal> result = new ArrayList<>();
            DBVal curVal = new DBVal(point, 0, 0);
            int expectedResultCount = (int) ((tsEnd - tsStart + step - 1) / step) + 1;
            long[] timestamps = new long[expectedResultCount];
            double[] values = new double[expectedResultCount];
            int curWritePos = 0;
            while (reader.hasNext()) {
                boolean shouldContinue = true;
                reader.next();
                double inValue = reader.getDouble(1);
                long inTimestamp = reader.getLong(2);
                long curWriteMaxTs = tsStart + curWritePos * step;
                while (inTimestamp > curWriteMaxTs) {
                    // Check if current value is valid
                    if (curWriteMaxTs - timestamps[curWritePos] > lookBackDuration) {
                        values[curWritePos] = VALUE_INVALID;
                    }
                    // Move next
                    curWritePos++;
                    if (curWritePos >= expectedResultCount) {
                        shouldContinue = false;
                        break;
                    }
                    curWriteMaxTs = tsStart + curWritePos * step;
                    // Copy previous value
                    timestamps[curWritePos] = timestamps[curWritePos - 1];
                    values[curWritePos] = values[curWritePos - 1];
                }
                if (!shouldContinue) {
                    break;
                }
                timestamps[curWritePos] = inTimestamp;
                values[curWritePos] = inValue;
            }
            while (curWritePos < expectedResultCount) {
                long curWriteMaxTs = tsStart + curWritePos * step;
                if (curWriteMaxTs - timestamps[curWritePos] > lookBackDuration) {
                    values[curWritePos] = VALUE_INVALID;
                }
                curWritePos++;
                if (curWritePos >= expectedResultCount) {
                    break;
                }
                // Copy previous value
                timestamps[curWritePos] = timestamps[curWritePos - 1];
                values[curWritePos] = values[curWritePos - 1];
            }

            return new PointVals(point, expectedResultCount, timestamps, values);
        } catch (Exception e) {
            LOG.error("selectPointValsFromDBWithStep error: ", e);
            return null;
        }
    }

    /// Read from ClickHouse
    ///
    /// 从数据库中依据给定的聚合方法，聚合指定时间范围内的数据
    protected double selectPointAggValueFromDB(Point point, long tsStart, long tsEnd, String aggMethod) {
        Map<String, Object> queryParams = new HashMap<>();
        queryParams.put("metricName", point.getMetricName());
        queryParams.put("pointName", point.getPointName());
        queryParams.put("tsStart", tsStart);
        queryParams.put("tsEnd", tsEnd);
        String sql = "SELECT " + aggMethod + "(value) FROM " + makeTableName() +
                " WHERE metric_name = {metricName:String}";
        if (StringUtils.isNotEmpty(point.getPointName())) {
            sql += " AND point_name = {pointName:String}";
        }
        sql += " AND nanoseconds >= {tsStart:Int64} AND nanoseconds <= {tsEnd:Int64}";
        try (QueryResponse response = client.query(sql, queryParams, new QuerySettings()).get()) {
            ClickHouseBinaryFormatReader reader = client.newBinaryFormatReader(response);
            // TODO: Clickhouse returns 0 for empty result?
            if (reader.hasNext()) {
                reader.next();
                return reader.getDouble(1);
            }
            return VALUE_INVALID;
        } catch (Exception e) {
            LOG.error("selectPointAggValueFromDB error: ", e);
            return VALUE_INVALID;
        }
    }

    /**
     * @date 被sendSinglePoint调用
     */
    protected int splitAndInsertSinglPoint(PointVals pointVals) {
        int insertedCount = 0;
        try {
            List<PointVals> pointValsList = splitPointVals(pointVals, this.splitSize);
            for (PointVals subPointVals : pointValsList) {
                this.insertPointValsToDB(subPointVals);
                insertedCount += subPointVals.getCount();
            }
        } catch (Exception e) {
            LOG.error("splitAndInsertSinglPoint error: ", e);
        }
        return insertedCount;

//        if (pointVals.isEmpty()) {
//            return 0;
//        }
//        int updateCount = 0;
//        try {
//            List<PointVals> pointValsList = splitPointVals(pointVals, this.splitSize);
//            for (PointVals subPointVals : pointValsList) {
//                SendRequest.histDataImport(host, insertPort, nameSpace, subPointVals, this.metricNamePrefix, this.getToken());
//                updateCount += subPointVals.getCount();
//            }
//        } catch (Exception e) {
//            LOG.error("splitAndInsertSinglPoint error: ", e);
//            return 0;
//        }
//        return updateCount;
    }

    public int insertSinglPoint(PointVals pointVals) {
        return this.splitAndInsertSinglPoint(pointVals);
    }

    /**
     * 写入多点的多条数据
     */
    public int insertMultiPoint(List<DBVal> dbValList) {
        return this.splitAndInsertMultiPoint(dbValList);
    }


    /**
     * getPointValueAtTime调用
     */
    protected DBVal getPointValueAtUtc(Point point, long utcTimeMilliSeconds) {
        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query?query=%s&time=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, point.calPointName(metricNamePrefix), utcTimeMilliSeconds / 1000.0);
        DBVal dbVal = null;
        try {
            dbVal = SendRequest.getDataAtUtc(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }

        if (dbVal == null) {
            dbVal = new DBVal(point, utcTimeMilliSeconds);
            dbVal.setValueStatusInValid();
        }
        return dbVal;
    }

    /**
     * getPointsValuesAtTime调用
     */
    protected Map<Point, DBVal> getPointValueListAtUtc(String metricName, HashMap<String, String> commonTagValuePairs, String tagNameWithDifferentValue, List<String> tagValues, long utcTimeMilliSeconds) {
        Map<Point, DBVal> result = new HashMap<>();

        String queryString = Point.calQueryStringForMultiPoint(metricNamePrefix, metricName, commonTagValuePairs, tagNameWithDifferentValue, tagValues);

        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query?query=%s&time=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, queryString, utcTimeMilliSeconds / 1000.0);
        try {
            result = SendRequest.getDatasAtUtc(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * getPointsValuesAtTime调用
     */
    protected Map<Point, DBVal> getPointValueListAtUtc(String metricName, HashMap<String, String> commonTagValuePairs, long utcTimeMilliSeconds) {
        Map<Point, DBVal> result = new HashMap<>();

        String queryString = Point.calQueryStringForMultiPoint(metricNamePrefix, metricName, commonTagValuePairs);

        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query?query=%s&time=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, queryString, utcTimeMilliSeconds / 1000.0);
        try {
            result = SendRequest.getDatasAtUtc(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * 被getRTValueList调用
     */
    protected List<DBVal> getPointValueListAtUtc(List<Point> points, long utcTimeMilliSeconds) {
        List<DBVal> result = new LinkedList<>();
        for (Point point : points) {
            result.add(this.getPointValueAtUtc(point, utcTimeMilliSeconds));
        }
        return result;
    }


    /**
     * 根据间隔取某个点的值 该方法将向前查找一定范围内(lookBack)的数据并填充，若向前查找无数据，则在该点设置无效值
     * Attention please, this method will cost a lot, if data sparsity exists with a small period.
     *
     * @param point
     * @param start
     * @param end
     * @param step
     * @return
     */
    protected List<DBVal> getDataInTimeRangeWithStepUseLookback(Point point, long start, long end, long step, long lookBack) {
        // Version Easy GET-IT.
        List<DBVal> dbValList = new LinkedList<>();
        List<DBVal> baseList = getDataInTimeRange(point, start - lookBack, end);
        if (baseList == null || baseList.isEmpty()) {
            return dbValList;
        }

        //handle fix with step
        long time = start;
        double value = 0;
        DBVal.ValueStatus status = DBVal.ValueStatus.INVALID;
        long lastTimestamp;
        if (baseList.get(baseList.size() - 1).getUtcTimeMilliSeconds() <= start) {
            value = baseList.get(baseList.size() - 1).getValue();
            status = baseList.get(baseList.size() - 1).getValueStatus();
            lastTimestamp = baseList.get(baseList.size() - 1).getUtcTimeMilliSeconds();
            lastTimestamp += lookBack;
            for (; time <= end; time += step) {
                if (time > lastTimestamp) {
                    dbValList.add(new DBVal(point, time, value, DBVal.ValueStatus.INVALID));
                } else {
                    dbValList.add(new DBVal(point, time, value, status));
                }
            }
            return dbValList;
        }

        //i代表当前index
        lastTimestamp = 0;
        for (int i = 0, size = baseList.size(); time <= end; time += step) {
            while (i < size && baseList.get(i).getUtcTimeMilliSeconds() <= time) {
                value = baseList.get(i).getValue();
                status = baseList.get(i).getValueStatus();
                lastTimestamp = baseList.get(i).getUtcTimeMilliSeconds();
                i++;
            }

            //i>=size或者i对应的时间大于time
            if (time - lastTimestamp > lookBack) {
                dbValList.add(new DBVal(point, time, value, DBVal.ValueStatus.INVALID));
            } else {
                dbValList.add(new DBVal(point, time, value, status));
            }
        }
        return dbValList;
    }

    /**
     *
     */
    protected PagedRangeQueryResult getDataInTimeRangeWithStepUseLookbackAndPage(Point point, long start, long end, long step, long lookBack, int pageIndex, int pageSize) throws Exception {
        // Version Easy GET-IT.
        List<DBVal> dbValList = new LinkedList<>();
        PagedRangeQueryResult pagedRangeQueryResult;

        if (point == null || pageSize <= 0 || pageIndex < 0 || step <= 0 || start > end || lookBack < 0)
            throw new Exception("point==null || pageSize<=0 || pageIndex<0 || step<=0 || start>end || lookBack<0");
        int pointTotalCount = this.countHistSnap(point, start, end, step, lookBack);

        if (pointTotalCount == 0) {
            pagedRangeQueryResult = new PagedRangeQueryResult(point, start, end, step, lookBack, dbValList, pageIndex, pageSize, 0);
            return pagedRangeQueryResult;
        }

        if (pageSize > pointTotalCount) pageSize = pointTotalCount;
        //页面数
        int pageCount = pointTotalCount % pageSize == 0 ? pointTotalCount / pageSize : pointTotalCount / pageSize + 1;

        if (pageIndex >= pageCount) {
            pagedRangeQueryResult = new PagedRangeQueryResult(point, start, end, step, lookBack, dbValList, pageIndex, pageSize, pageCount);
            pagedRangeQueryResult.setSuccess(false);
            return pagedRangeQueryResult;
        }

        long pageStartTime = start + (pageIndex * pageSize) * step;
        long pageEndTime = start + (pageIndex * pageSize + pageSize - 1) * step;

        if (pageEndTime > end) pageEndTime = end;
//        dbValList = this.getDataInTimeRangeWithStepUseLookback(point, pageStartTime, pageEndTime, step, lookBack);
        dbValList = this.getHistSnap(point, pageStartTime, pageEndTime, step, lookBack).getDbValList();
        return new PagedRangeQueryResult(point, start, end, step, this.lookBack, dbValList, pageIndex, pageSize, pageCount);
    }

    /**
     * 被getHistSnap调用
     */
    protected List<DBVal> getDataInTimeRangeWithStep(Point point, long start, long end, long step) {
        List<DBVal> dbValList = new ArrayList<>();
        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query_range?query=%s&start=%f&end=%f&step=%dms", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, point.calPointName(this.metricNamePrefix), start / 1000.0, end / 1000.0, step);
        try {
            dbValList = SendRequest.getDataInTimeRange(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return dbValList;
    }

    /**
     * 被getHistSnap调用
     */
    public List<DBVal> getHistDataInTimeRange(Point point, long start, long end) {
        List<DBVal> dbValList = new ArrayList<>();
        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query_range?query=%s&start=%f&end=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, point.calPointName(this.metricNamePrefix), start / 1000.0, end / 1000.0);
        try {
            dbValList = SendRequest.getDataInTimeRange(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return dbValList;
    }

    /**
     * 被getHistRaw调用
     */
    protected List<DBVal> getDataInTimeRange(Point point, long start, long end) {
        List<DBVal> dbValList = new ArrayList<>();
        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/export?match[]=%s&start=%f&end=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, point.calMatchPointName(this.metricNamePrefix), start / 1000.0, end / 1000.0);

        try {
            dbValList = SendRequest.getDataInTimeRangeRaw(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return dbValList;
    }

    public DBVal getPointValueAtTime(Point point, long time) {
        return this.selectPointValsFromDB(point, 0, time, true, 1)
                .getDBVal(0);
//        return this.getPointValueAtUtc(point, time);
    }

    public List<DBVal> getPointsValuesAtTime(List<Point> pointList, long time) {
        return this.selectManyPointValsFromDB(pointList, 0, time, true, 1)
                .stream()
                .map(x -> x.getDBVal(0))
                .collect(Collectors.toList());
//        return this.getPointValueListAtUtc(pointList, time);
    }

    public Map<Point, DBVal> getPointsValuesAtTime(String metricName, HashMap<String, String> commonTagValuePairs, String tagNameWithDifferentValue, List<String> tagValues, long time) {
        // TODO...
        throw new UnsupportedOperationException("Not implemented yet!");
//        return this.getPointValueListAtUtc(metricName, commonTagValuePairs, tagNameWithDifferentValue, tagValues, time);
    }

    public Map<Point, DBVal> getPointsValuesAtTime(String metricName, HashMap<String, String> commonTagValuePairs, long time) {
        // TODO...
        throw new UnsupportedOperationException("Not implemented yet!");
//        return this.getPointValueListAtUtc(metricName, commonTagValuePairs, time);
    }

    public DBVal getRTValue(Point point) {
        return this.selectPointValsFromDB(point, 0, Long.MAX_VALUE, true, 1)
                .getDBVal(0);

//        DBVal dbVal = null;
//        long currentTimeMillis = System.currentTimeMillis();
//        dbVal = getPointValueAtUtc(point, currentTimeMillis);
//        if (dbVal == null) {
//            dbVal = new DBVal(point, currentTimeMillis);
//            dbVal.setValueStatusInValid();
//        }
//        return dbVal;
    }

    public List<DBVal> getRTValueList(List<Point> points) {
        return this.selectManyPointValsFromDB(points, 0, Long.MAX_VALUE, true, 1)
                .stream()
                .map(x -> x.getDBVal(0))
                .collect(Collectors.toList());

//        List<DBVal> dbValList;
//        dbValList = getPointValueListAtUtc(points, System.currentTimeMillis());
//        return dbValList;
    }

    public Map<Point, DBVal> getRTValueList(String metricName, HashMap<String, String> commonTagValuePairs, String tagNameWithDifferentValue, List<String> tagValues) {
        // TODO...
        throw new UnsupportedOperationException("Not implemented yet!");

//        long time = System.currentTimeMillis();
//        return this.getPointValueListAtUtc(metricName, commonTagValuePairs, tagNameWithDifferentValue, tagValues, time);
    }

    public Map<Point, DBVal> getRTValueList(String metricName, HashMap<String, String> commonTagValuePairs) {
        // TODO...
        throw new UnsupportedOperationException("Not implemented yet!");

//        long time = System.currentTimeMillis();
//        return this.getPointValueListAtUtc(metricName, commonTagValuePairs, time);
    }

    /**
     * 获取单点历史数据，指定步长
     */
    public RangeQueryResult getHistSnap(Point point, long start, long end, long step) {
        return this.getHistSnap(point, start, end, step, this.lookBack);

//        if (point == null || step <= 0 || start > end) {
//            RangeQueryResult result = new RangeQueryResult(point, start, end, step, this.lookBack, new ArrayList<>());
//            result.setSuccess(false);
//            return result;
//        }
//
//        List<DBVal> dbValList;
//        dbValList = getDataInTimeRangeWithStepUseLookback(point, start, end, step, this.lookBack);
//        return new RangeQueryResult(point, start, end, step, this.lookBack, dbValList);
    }

    public List<DBVal> getHistSnapV1(Point point, long start, long end, long step) {
        RangeQueryResult rangeQueryResult = this.getHistSnap(point, start, end, step);
        return rangeQueryResult.getDbValList();
    }

    /**
     * 获取单点历史数据，指定步长
     */
    public RangeQueryResult getHistSnap(Point point, long start, long end, long step, long lookBack) {
        if (point == null || step < 0 || start > end) {
            RangeQueryResult result = new RangeQueryResult(point, start, end, step, lookBack, new ArrayList<>());
            result.setSuccess(false);
            return result;
        }

        if (step == 0) {
            PointVals pointVals = this.selectPointValsFromDB(point, start, end, false, 0);
            return new RangeQueryResult(point, start, end, step, lookBack, pointVals.getDBValList());
        }

        PointVals pointVals = this.selectPointValsFromDBWithStep(point, start, end, step, lookBack);
        List<DBVal> dbValList = pointVals.getDBValList();
        for (DBVal dbVal : dbValList) {
            if (Double.isNaN(dbVal.getValue())) {
                dbVal.setValueStatusInValid();
            }
        }
        return new RangeQueryResult(point, start, end, step, lookBack, dbValList);

//        if (point == null || step <= 0 || start > end || lookBack < 0) {
//            RangeQueryResult result = new RangeQueryResult(point, start, end, step, this.lookBack, new ArrayList<>());
//            result.setSuccess(false);
//            return result;
//        }
//
//        List<DBVal> dbValList;
////        dbValList = getDataInTimeRangeWithStep(point, start, end, step);
//        dbValList = getDataInTimeRangeWithStepUseLookback(point, start, end, step, lookBack);
//        return new RangeQueryResult(point, start, end, step, this.lookBack, dbValList);
    }

    /**
     * 获取单点历史数据，指定步长、分页
     */
    public PagedRangeQueryResult getHistSnap(Point point, long start, long end, long step, int pageIndex, int pageSize) {
        PagedRangeQueryResult pagedRangeQueryResult = null;
        try {
            pagedRangeQueryResult = getDataInTimeRangeWithStepUseLookbackAndPage(point, start, end, step, this.lookBack, pageIndex, pageSize);
        } catch (Exception e) {
            LOG.error("getHistSnap error: ", e);
            pagedRangeQueryResult = new PagedRangeQueryResult(point, start, end, step, this.lookBack, new ArrayList<>(), pageIndex, pageSize, 0);
            pagedRangeQueryResult.setSuccess(false);
        }
        return pagedRangeQueryResult;
    }

    /**
     * 获取单点历史数据，指定步长、分页
     */
    public PagedRangeQueryResult getHistSnap(Point point, long start, long end, long step, long lookBack, int pageIndex, int pageSize) {
        PagedRangeQueryResult pagedRangeQueryResult = null;
        try {
            pagedRangeQueryResult = getDataInTimeRangeWithStepUseLookbackAndPage(point, start, end, step, lookBack, pageIndex, pageSize);
        } catch (Exception e) {
            LOG.error("getHistSnap error: ", e);
            pagedRangeQueryResult = new PagedRangeQueryResult(point, start, end, step, lookBack, new ArrayList<>(), pageIndex, pageSize, 0);
            pagedRangeQueryResult.setSuccess(false);
        }
        return pagedRangeQueryResult;
    }


    /**
     * 计算单点指定步长数据的条数
     */
    public int countHistSnap(Point point, long start, long end, long step, long lookBack) {
        return this.countHistSnapReal(point, start, end, step, lookBack);
    }

    /**
     * 计算单点指定步长数据的条数
     */
    public int countHistSnap(Point point, long start, long end, long step) {
        return this.countHistSnapReal(point, start, end, step, this.lookBack);
    }

    protected int countHistSnapReal(Point point, long start, long end, long step, long lookBack) {
        if (end < start || step <= 0) return 0;

        Long countOverTime = countOverTime(point, start - lookBack, end);
        if (countOverTime == null) {
            return 0;
        }

        int count = (int) ((end - start + step - 1) / step) + 1;
        return count;
    }

    /**
     * 获取多点历史数据，指定步长
     */
    public Map<Point, RangeQueryResult> getHistSnap(List<Point> pointList, long start, long end, long step) {
        Map<Point, RangeQueryResult> resultMap = new HashMap<>();
        if (pointList == null || pointList.isEmpty()) {
            return resultMap;
        }

        for (Point point : pointList) {
            RangeQueryResult result = this.getHistSnap(point, start, end, step, this.lookBack);
            resultMap.put(point, result);
        }

        return resultMap;
    }

    /**
     * 获取多点历史数据，指定步长
     */
    public Map<Point, RangeQueryResult> getHistSnap(List<Point> pointList, long start, long end, long step, long lookBack) {
        Map<Point, RangeQueryResult> resultMap = new HashMap<>();
        if (pointList == null || pointList.isEmpty()) {
            return resultMap;
        }

        for (Point point : pointList) {
            RangeQueryResult result = this.getHistSnap(point, start, end, step, lookBack);
            resultMap.put(point, result);
        }

        return resultMap;
    }

    /**
     * 获取多点历史数据，指定步长、分页
     */
    public Map<Point, PagedRangeQueryResult> getHistSnap(List<Point> pointList, long start, long end, long step, int pageIndex, int pageSize) {
        Map<Point, PagedRangeQueryResult> resultMap = new HashMap<>();
        if (pointList == null || pointList.isEmpty()) {
            return resultMap;
        }

        for (Point point : pointList) {
            PagedRangeQueryResult result = this.getHistSnap(point, start, end, step, this.lookBack, pageIndex, pageSize);
            resultMap.put(point, result);
        }

        return resultMap;
    }

    /**
     * 获取多点历史数据，指定步长、分页
     */
    public Map<Point, PagedRangeQueryResult> getHistSnap(List<Point> pointList, long start, long end, long step, long lookBack, int pageIndex, int pageSize) {
        Map<Point, PagedRangeQueryResult> resultMap = new HashMap<>();
        if (pointList == null || pointList.isEmpty()) {
            return resultMap;
        }

        for (Point point : pointList) {
            PagedRangeQueryResult result = this.getHistSnap(point, start, end, step, lookBack, pageIndex, pageSize);
            resultMap.put(point, result);
        }

        return resultMap;
    }


    /**
     * 计算单点指定步长数据的条数
     */
    public Map<Point, Integer> countHistSnap(List<Point> pointList, long start, long end, long step, long lookBack) {
        return this.countHistSnapReal(pointList, start, end, step, lookBack);
    }

    public Map<Point, Integer> countHistSnap(List<Point> pointList, long start, long end, long step) {
        return this.countHistSnapReal(pointList, start, end, step, this.lookBack);
    }

    public Map<Point, Integer> countHistSnapReal(List<Point> pointList, long start, long end, long step, long lookBack) {
        Map<Point, Integer> result = new HashMap<>();
        if (pointList == null || pointList.isEmpty()) {
            return result;
        }

        for (Point point : pointList) {
            int count = this.countHistSnap(point, start, end, step);
            result.put(point, count);
        }

        return result;
    }

    /**
     * 获取多点历史数据，指定步长
     */
    public Map<Point, List<DBVal>> getHistSnap(String metricName, HashMap<String, String> commonTagValuePairs, String tagNameWithDifferentValue, List<String> tagValues, long start, long end, long step) {
        List<Point> points = new ArrayList<>();
        if (StringUtils.isEmpty(metricName) || StringUtils.isEmpty(tagNameWithDifferentValue) || tagValues == null || tagValues.isEmpty()) {
            return new HashMap<>();
        }
        if (!tagNameWithDifferentValue.equals("pointName")) {
            points.add(new Point(metricName, commonTagValuePairs));
        } else {
            for (String tagValue : tagValues) {
                HashMap<String, String> tagValuePairs = new HashMap<>(commonTagValuePairs);
                tagValuePairs.put(tagNameWithDifferentValue, tagValue);
                points.add(new Point(metricName, tagValuePairs));
            }
        }
        Map<Point, RangeQueryResult> resultMap = this.getHistSnap(points, start, end, step);
        return resultMap.entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, x -> x.getValue().getDbValList()));

//        Map<Point, List<DBVal>> dbValMapListList = new HashMap<>();
//        if (metricName == null) {
//            return dbValMapListList;
//        }
//
//        String queryString = Point.calQueryStringForMultiPoint(metricNamePrefix, metricName, commonTagValuePairs, tagNameWithDifferentValue, tagValues);
//        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query_range?query=%s&start=%f&end=%f&step=%dms", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, queryString, start / 1000.0, end / 1000.0, step);
//
//        try {
//            dbValMapListList = SendRequest.getDataInTimeRangeMultiPoint(queryUrl, this.getToken());
//        } catch (Exception e) {
//            e.printStackTrace();
//        }
//        return dbValMapListList;
    }

    /**
     * 查询单点时间段内所有值
     */
    public List<DBVal> getHistRaw(Point point, long start, long end) {
        return this.selectPointValsFromDB(point, start, end, false, 0)
                .getDBValList();
//        List<DBVal> dbValList;
//        dbValList = getDataInTimeRange(point, start, end);
//        return dbValList;
    }

    /**
     * 查询多点时间段内所有值
     */
    public Map<Point, List<DBVal>> getHistRaw(List<Point> pointList, long start, long end) {
        return this.selectManyPointValsFromDB(pointList, start, end, false, 0)
                .stream()
                .collect(Collectors.toMap(PointVals::getPoint, PointVals::getDBValList));
//        Map<Point, List<DBVal>> dbValMapListList = new HashMap<>();
//        if (pointList == null || pointList.isEmpty()) {
//            return dbValMapListList;
//        }
//        dbValMapListList = getDataAtTimeRangeMap(pointList, start, end);
//        return dbValMapListList;
    }

    /**
     * 被getHistRaw调用
     */
    protected Map<Point, List<DBVal>> getDataAtTimeRangeMap(List<Point> pointList, long start, long end) {
        Map<Point, List<DBVal>> dbValMapList = new HashMap<>();
        String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/export", this.host, this.selectPort, this.nameSpace, Constants.ROUTE_PREFIX);

        try {
            dbValMapList = SendRequest.getDataInTimeRangeRawMap(queryUrl, this.metricNamePrefix, pointList, start, end, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
        }
        return dbValMapList;
    }

    public DBVal getMinValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "min", startTime, endTime);
    }

    public DBVal getMedianValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "median", startTime, endTime);
    }

    public DBVal getStddevValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "stddevPop", startTime, endTime);
    }

    public DBVal getStdvarValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "varPop", startTime, endTime);
    }

    public DBVal getSumValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "sum", startTime, endTime);
    }

    public DBVal calAggOverTime(Point point, String aggMethod, long startTime, long endTime) {
        double value = this.selectPointAggValueFromDB(point, startTime, endTime, aggMethod);
        DBVal dbVal = new DBVal(point, endTime, value);
        if (Double.isNaN(value)) {
            dbVal.setValueStatusInValid();
        }
        return dbVal;

//        if (startTime > endTime) {
//            return null;
//        }
//        DBVal dbVal = null;
//        try {
//            long utcTime = System.currentTimeMillis();
//            long offset = utcTime - endTime;
//            long timeRange = endTime - startTime;
//
//            String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query?query=%s(%s[%dms])&time=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, aggMethod, point.calPointName(this.metricNamePrefix), timeRange, endTime / 1000.0);
////            System.out.println(String.format("querUrl:%s", queryUrl));
////            String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query?query=%s(%s[%dms] offset %dms)",
////                    host, selectPort, nameSpace, Constants.ROUTE_PREFIX, aggMethod,
////                    point.calPointName(this.metricNamePrefix), timeRange, offset);
//            queryUrl = queryUrl.replace(" ", "%20");
//            dbVal = SendRequest.getDataAtUtc(queryUrl, this.getToken()); // equal to return DBval
//            if (dbVal == null) {
//                dbVal = new DBVal();
//            }
//        } catch (Exception e) {
//            e.printStackTrace();
//
//        }
//        return dbVal;
    }

    public Long countOverTime(Point point, long start, long end) {
        double result = this.selectPointAggValueFromDB(point, start, end, "count");
        if (Double.isNaN(result)) {
            return null;
        }
        return (long) result;

//        DBVal dbVal = this.aggOverTime(point, start, end, Constants.COUNT);
//        if (dbVal == null || !dbVal.isValueValid()) return null;
//        else {
//            return dbVal.getLongValue();
//        }
    }

    public DBVal aggOverTime(Point point, long start, long end, String aggMethod) {

        if (start > end) {
            DBVal val = new DBVal();
            val.setValue(0);
            return val;
        }

        try {
//            long timeRange = (end - start)/1000;
            long timeRange = (end - start);

//            String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query_range?query=sum(%s_over_time(%s[%ds]))&start=%d&end=%d&step=%ds",
//                    host, selectPort, nameSpace, Constants.ROUTE_PREFIX, aggMethod, point.calPointName(this.metricNamePrefix),
//                    timeRange,start/1000,end/1000, timeRange);
            String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/query_range?query=sum(%s_over_time(%s[%dms]))&start=%f&end=%f&step=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, aggMethod, point.calPointName(this.metricNamePrefix), timeRange, start / 1000.0, end / 1000.0, timeRange / 1000.0);
            queryUrl = queryUrl.replace("[", "%5B").replace("]", "%5D").replace(" ", "%20").replace("{", "%7B").replace("}", "%7D").replace("\"", "%22");

            HttpURLConnection conn = (HttpURLConnection) (new URL(queryUrl)).openConnection();
            this.configHttpHeaderToken(conn);
            conn.setConnectTimeout(5000);
            conn.setRequestMethod("GET");
            if (conn.getResponseCode() != 200) {
//                log.error("metricName:{} tagNameValueSetMap:{} start:{} end:{}  now:{} get response code:{}",
//                        metricName, tagNameValueSetMap, start, end, now, conn.getResponseCode() );
                DBVal val = new DBVal();
                val.setValue(0);
                val.setValueStatusInValid();
                return val;
            }

            InputStream inputStream = conn.getInputStream();
            InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
            BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
            String str = null;
            while ((str = bufferedReader.readLine()) != null) {
                JSONObject res = JSONObject.parseObject(str);
                JSONArray result = res.getJSONObject("data").getJSONArray("result");
                if (result.size() == 0) {

//                    log.error("metricName:{} tagNameValueSetMap:{} start:{} end:{}  now:{} get response result.size():0",
//                            metricName, tagNameValueSetMap, start, end, now);
                    bufferedReader.close();
                    conn.disconnect();
                    DBVal val = new DBVal();
                    val.setValue(0);
                    val.setValueStatusInValid();
                    return val;
                }

                if (result.size() > 1) {
//                    log.error("metricName:{} tagNameValueSetMap:{} start:{} end:{}  now:{} get response result.size()>1",
//                            metricName, tagNameValueSetMap, start, end, now);
                }

                int index = 0;
                JSONObject object = result.getJSONObject(index);
                JSONObject metricInfo = object.getJSONObject("metric");
                Point pointResult = Point.fromJSONObject(metricInfo);
                JSONArray values = object.getJSONArray("values");
                DBVal dbVal = new DBVal();
                JSONArray value = null;
                dbVal.setPoint(pointResult);
                if (values.size() > 0) {
                    value = values.getJSONArray(values.size() - 1);
                    long timeStamp = value.getBigDecimal(0).multiply(new BigDecimal(1000)).longValue();
                    if (timeStamp != end) {
                        dbVal.setUtcTimeMilliSeconds(end);
                        dbVal.setValue(0);
                        dbVal.setValueStatusInValid();
                    } else {
                        dbVal.setUtcTimeMilliSeconds(timeStamp);
                        dbVal.setValue(value.getDoubleValue(1));
                    }
                } else {
                    dbVal.setUtcTimeMilliSeconds(end);
                    dbVal.setValue(0);
                    dbVal.setValueStatusInValid();
                }
                bufferedReader.close();
                conn.disconnect();
                return dbVal;
            }

        } catch (Exception e) {
            e.printStackTrace();
            DBVal val = new DBVal();
            val.setValue(0);
            val.setValueStatusInValid();
            return val;
        }
        DBVal val = new DBVal();
        val.setValue(0);
        val.setValueStatusInValid();
        return val;
    }


    public DBVal getMaxValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "max", startTime, endTime);
    }

    public DBVal getAvgValue(Point point, long startTime, long endTime) {
        return calAggOverTime(point, "avg", startTime, endTime);
    }

    protected int getSplitSize() {
        return splitSize;
    }

    protected void setSplitSize(int splitSize) {
        this.splitSize = splitSize;
    }

    protected int getOneThreadDataSize() {
        return oneThreadDataSize;
    }

    protected void setOneThreadDataSize(int oneThreadDataSize) {
        this.oneThreadDataSize = oneThreadDataSize;
    }

    //admin
    public String getStatus() {
        String response = "";
        try {
            String url = String.format("http://%s:%d/select/%d/%s/api/v1/status/tsdb", host, selectPort, nameSpace, Constants.ROUTE_PREFIX);
            url = url.replace(" ", "%20");
            response = SendRequest.getResponse(url, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();

        }
        return response;

    }

    public Set<Object> tagValues(String metricName, HashMap<String, String> filterTagValuePairs, String tagName) {
        Set<Object> result;
        long end = System.currentTimeMillis();
        long start = end - 10L * 365 * 24 * 3600 * 1000;

        try {
            Point point = new Point(metricName, filterTagValuePairs);
            String queryUrl = String.format("http://%s:%d/select/%d/%s/api/v1/label/%s/values?match[]=%s&start=%f&end=%f", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, tagName, point.calPointName(this.metricNamePrefix), start / 1000.0, end / 1000.0);
            queryUrl = queryUrl.replace("[", "%5B").replace("]", "%5D").replace(" ", "%20").replace("{", "%7B").replace("}", "%7D").replace("\"", "%22");
            result = SendRequest.labelValues(queryUrl, this.getToken());
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        return result;
    }

    public Set<Object> tagValues(String metricName, String tagName) {
        HashMap<String, String> filterTagValuePairs = new HashMap<>();
        return tagValues(metricName, filterTagValuePairs, tagName);
    }

    public RangePagingQuery createPagingQuery(Point point, long startTime, long endTime, int pageSize) throws Exception {
        if (pageSize <= 0) throw new Exception("pageSize <= 0");
        List<DBVal> result = getHistRaw(point, startTime, endTime);

        if (result.size() == 0) {
            return new RangePagingQuery(startTime, endTime, pageSize, 0, result);
        }

        int pageCount = result.size() % pageSize == 0 ? result.size() / pageSize : result.size() / pageSize + 1;

        return new RangePagingQuery(startTime, endTime, pageSize, pageCount, result);
    }

    public MultiPointRangePagingQuery createMultiPointPagingQuery(List<Point> pointList, long startTime, long endTime, int pageSize) throws Exception {
        if (pageSize <= 0) throw new Exception("pageSize <= 0");
        if (pointList.isEmpty()) throw new Exception("pointList isEmpty");
        Map<Point, RangePagingQuery> rangePagingQueryMap = new HashMap<>();

        for (int i = 0; i < pointList.size(); i++) {
            Point point = pointList.get(i);
            RangePagingQuery rangePagingQuery = createPagingQuery(point, startTime, endTime, pageSize);
            rangePagingQueryMap.put(point, rangePagingQuery);
        }
        MultiPointRangePagingQuery result = new MultiPointRangePagingQuery(startTime, endTime, pageSize, rangePagingQueryMap);

        return result;
    }


    /**
     * 本方法用于数据删除,需要搭配forceMerge和getForceMergePath使用,才能完整实现表面与存储空间上的彻底删除
     */
    public boolean deletePoint(Point point) {
        try {
            String deleteUrl = String.format("http://%s:%d/delete/%d/%s/api/v1/admin/tsdb/delete_series?match[]=%s", host, selectPort, nameSpace, Constants.ROUTE_PREFIX, point.calMatchPointName(this.metricNamePrefix));

            HttpURLConnection conn = (HttpURLConnection) new URL(deleteUrl).openConnection();
            conn.setConnectTimeout(5000);
            conn.setRequestMethod("GET");
            if (conn.getResponseCode() != 204) {
                return false;
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /**
     * 本方法用于数据删除
     *
     * @param pathName String类型
     * @return boolean
     * @author Hu
     * @date 2020-11-05 00:00
     */
    protected boolean forceMerge(String pathName) {
        try {
            String commandUrl = "http://" + host + ":" + port + "/internal/force_merge?partition_prefix=" + pathName;
            HttpURLConnection conn = (HttpURLConnection) new URL(commandUrl).openConnection();
            conn.setConnectTimeout(5000);
            conn.setRequestMethod("GET");
            if (conn.getResponseCode() != 204) {
                return false;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        return true;
    }

}
